package com.sillycycle.bagleyd.bevel;

/*
 * @(#)Bevel2D.java
 *
 * Copyright 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;

/**
 * The <code>Bevel2D</code> class does all the 2D graphics for
 * <code>BevelFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Bevel2D extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int FRONT_LOC = 0;
	static final int BACK_LOC = 11;
	static final int COORD2D = 8;
	static final int MAXX = 3;
	static final int MAXY = 4;
	static final int MAXXY = 4; /* (MAX(MAXX,MAXY)) */
	static final int MAX_RECT = 12; /* (MAXX*MAXY) */

	Point[][] innerTriangleUnit = new Point[Bevel.MAX_ORIENT][4];
	Point[][] outerTriangleUnit = new Point[Bevel.MAX_ORIENT][4];
	Point[] letterUnit = new Point[Bevel.MAX_CUBES];
	int[][] innerTriangleListX = new int[Bevel.MAX_ORIENT][4];
	int[][] innerTriangleListY = new int[Bevel.MAX_ORIENT][4];
	int[][] outerTriangleListX = new int[Bevel.MAX_ORIENT][4];
	int[][] outerTriangleListY = new int[Bevel.MAX_ORIENT][4];
	Point[] letterList = new Point[Bevel.MAX_CUBES];
	Point[] innerOffsetList = new Point[4];
	Point[] outerOffsetList = new Point[4];
	static final int[/*MAX_RECT*/] planeToCube =
		{6, 0, 6, 1, 2, 3, 6, 4, 6, 6, 5, 6};
	static int[/*MAX_FACES*/]  cubeToPlane =
		{1, 3, 4, 5, 7, 10};

	Point pos = new Point(0, 0);
	Point puzzleSize, coreSize, puzzleOffset, letterOffset;
	int cubeLength, delta, orientLineLength;
	int viewLength, faceLength, diamondLength;
	int lastX, lastY;

	JFrame      frame = null;
	Bevel       bevel = null;
	int         randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false, vertical = false;

	public Bevel2D(JFrame frame, Bevel bevel) {
		this.frame = frame;
		this.bevel = bevel;
		firstPaint = true;
	}

	public boolean checkStarted() {
		return bevel.checkStarted();
	}

	void drawOrientLine(Graphics g, int face, int rotation,
			int dx, int dy) {
		int x1 = dx, x2 = dx, y1 = dy, y2 = dy;

		switch (rotation) {
		case 0:
			y1 -= delta;
			y2 -= (delta + orientLineLength / 2);
			break;
		case 1:
			x1 += 1;
			x2 += 1 + orientLineLength / 2;
			break;
		case 2:
			y1 += delta;
			y2 += delta + orientLineLength / 2;
			break;
		case 3:
			x1 -= 1;
			x2 -= (1 + orientLineLength / 2);
			break;
		default:
			System.out.println("drawOrientLine: rotation " +
				rotation);
		}
		g.drawLine(x1, y1, x2, y2);
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawInnerTriangle(int face, int position, int offsetPosition) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			int rotation, l, newCorner;
			Point start = new Point(0, 0);
			Color fillColor, lineColor;

			rotation = bevel.cubeLoc[face][position + Bevel.MAX_ORIENT].rotation;
			if (vertical || face != Bevel.MAX_FACES - 1) {
				start.x = (cubeToPlane[face] % MAXX) *
					viewLength + delta - 1;
				start.y = (cubeToPlane[face] / MAXX) *
					viewLength + delta - 1;
				newCorner = position;
			} else {
				start.x = (cubeToPlane[face] / MAXX) *
					viewLength + delta - 1;
				start.y = (cubeToPlane[face] % MAXX) *
					viewLength + delta - 1;
				newCorner = (position + Bevel.HALF) % Bevel.STRT;
				rotation = (rotation + Bevel.HALF) % Bevel.STRT;
			}
			start.x += puzzleOffset.x + delta;
			start.y += puzzleOffset.y + delta;
			l = bevel.cubeLoc[face][position + Bevel.MAX_ORIENT].face;
			if (offsetPosition == 0) {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = bevel.faceColor[l];
					lineColor = bevel.borderColor;
				}
			} else {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = bevel.faceColor[l];
					fillColor = bevel.borderColor;
				}
			}
			fillPolygon(g, fillColor, lineColor,
				start.x, start.y,
				innerTriangleListX[newCorner],
				innerTriangleListY[newCorner], 3);
			if (bevel.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = diamondLength / 10;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(bevel.faceNames.charAt(l));
				g.drawString(buf,
					start.x + letterList[newCorner + Bevel.MAX_ORIENT].x,
					start.y + letterList[newCorner + Bevel.MAX_ORIENT].y);
			}
			if (bevel.orientChoice) {
				g.setColor(lineColor);
				drawOrientLine(g, newCorner, rotation,
					start.x + letterList[newCorner + Bevel.MAX_ORIENT].x,
					start.y + letterList[newCorner + Bevel.MAX_ORIENT].y - 5);
			}
		} finally {
			g.dispose();
		}
	}

	void drawOuterTriangle(int face, int position, int offsetPosition) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			int rotation, l, newCorner;
			Point start = new Point(0, 0);
			Color fillColor, lineColor;

			rotation = bevel.cubeLoc[face][position].rotation;
			if (vertical || face != Bevel.MAX_FACES - 1) {
				start.x = (cubeToPlane[face] % MAXX) *
					viewLength + delta - 1;
				start.y = (cubeToPlane[face] / MAXX) *
					viewLength + delta - 1;
				newCorner = position;
			} else {
				start.x = (cubeToPlane[face] / MAXX) *
					viewLength + delta - 1;
				start.y = (cubeToPlane[face] % MAXX) *
					viewLength + delta - 1;
				newCorner = (position + Bevel.HALF) % Bevel.STRT;
				rotation = (rotation + Bevel.HALF) % Bevel.STRT;
			}
			start.x += puzzleOffset.x + delta;
			start.y += puzzleOffset.y + delta;
			l = bevel.cubeLoc[face][position].face;
			if (offsetPosition == 0) {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = bevel.faceColor[l];
					lineColor = bevel.borderColor;
				}
			} else {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = bevel.faceColor[l];
					fillColor = bevel.borderColor;
				}
			}
			fillPolygon(g, fillColor, lineColor,
				start.x, start.y,
				outerTriangleListX[newCorner],
				outerTriangleListY[newCorner], 3);
			if (bevel.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = diamondLength / 10;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(bevel.faceNames.charAt(l));
				g.drawString(buf,
					start.x + letterList[newCorner].x,
					start.y + letterList[newCorner].y);
			}
			if (bevel.orientChoice) {
				g.setColor(lineColor);
				drawOrientLine(g, newCorner, rotation,
					start.x + letterList[newCorner].x,
					start.y + letterList[newCorner].y - 5);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < Bevel.MAX_ORIENT; position++) {
			drawInnerTriangle(face, position, 0);
			drawOuterTriangle(face, position, 0);
		}
	}

	void drawAllPieces() {
		for (int face = 0; face < Bevel.MAX_FACES; face++)
			drawFace(face);
	}

	void narrowSelection() {
		// sets bevel.currentFace, bevel.currentPosition, bevel.currentDirection
		if (bevel.currentFace == Bevel.MAX_FACES - 1 &&
				bevel.currentDirection < Bevel.MAX_ORIENT && !vertical)
			if (bevel.currentDirection < Bevel.MAX_ORIENT)
				bevel.currentDirection = (bevel.currentDirection +
					Bevel.HALF) % Bevel.MAX_ORIENT;
			else if (bevel.currentDirection >= 2 * Bevel.MAX_ORIENT)
				bevel.currentDirection = 2 * Bevel.MAX_ORIENT +
					(bevel.currentDirection + Bevel.HALF) %
					Bevel.MAX_ORIENT;
	}

	int positionToPiece(int positionX, int positionY) {
		// sets bevel.currentPosition
		int faceX, faceY, i, j, face;
		int x = positionX - puzzleOffset.x;
		int y = positionY - puzzleOffset.y;

		faceX = x / viewLength;
		faceY = y / viewLength;
		i = x - faceX * viewLength;
		j = y - faceY * viewLength;

		if ((i > viewLength / 2) &&
				(j <= viewLength / 2 + 1))
			bevel.currentPosition = 0;
		else if ((i > viewLength / 2) &&
				(j > viewLength / 2))
			bevel.currentPosition = 1;
		else if ((i <= viewLength / 2 + 1) &&
				(j > viewLength / 2))
			bevel.currentPosition = 2;
		else if ((i <= viewLength / 2 + 1) &&
				(j <= viewLength / 2 + 1))
			bevel.currentPosition = 3;
		else
			return -1;
		if ((faceX != 1 && faceY != 1) ||
				(faceX >= 3 && vertical) ||
				(faceY >= 3 && !vertical))
			return -1;
		face = planeToCube[faceX + faceY * MAXX];
		if (faceX == 3) {
			face = Bevel.MAX_FACES - 1;
			if (bevel.currentPosition != 2 * Bevel.MAX_ORIENT)
				bevel.currentPosition = (bevel.currentPosition +
					Bevel.HALF) % Bevel.MAX_ORIENT;
		}
		/*if (i - j > viewLength / 2 - 3)
			bevel.currentPosition = 0;
		else if (i + j > 3 * viewLength / 2)
			bevel.currentPosition = 1;
		else if (j - i > viewLength / 2 - 2)
			bevel.currentPosition = 2;
		else if (i + j < viewLength / 2 + 7)
			bevel.currentPosition = 3;*/
		return face;
	}

	void resizePieces() {
		orientLineLength = diamondLength / 8;
		letterOffset = new Point(-2, 4);
		for (int i = 0; i < Bevel.MAX_CUBES; i++)
			letterList[i] = new Point(0, 0);
		for (int i = 0; i < Bevel.MAX_ORIENT; i++) {
			innerOffsetList[i] = new Point(0, 0);
			outerOffsetList[i] = new Point(0, 0);
		}
		for (int i = 0; i < Bevel.MAX_ORIENT; i++) {
			int delta1 = 2;
			switch (innerTriangleUnit[i][0].x) {
			case 0:
				innerOffsetList[i].x = 0;
				break;
			case 1:
				innerOffsetList[i].x = diamondLength / 2 -
					delta - 1;
				break;
			case 2:
				innerOffsetList[i].x = diamondLength / 2;
				break;
			case 3:
				innerOffsetList[i].x = diamondLength - delta - 1;
				break;
			default:
				break;
			}
			switch (innerTriangleUnit[i][0].y) {
			case 0:
				innerOffsetList[i].y = 0;
				break;
			case 1:
				innerOffsetList[i].y = diamondLength / 2 -
					delta - 1;
				break;
			case 2:
				innerOffsetList[i].y = diamondLength / 2 - 1;
				break;
			case 3:
				innerOffsetList[i].y = diamondLength - delta - 2;
				break;
			default:
				break;
			}

			switch (outerTriangleUnit[i][0].x) {
			case 0:
				outerOffsetList[i].x = -2;
				break;
			case 1:
				outerOffsetList[i].x = diamondLength / 2 -
					delta - 1;
				break;
			case 2:
				outerOffsetList[i].x = diamondLength / 2 + 1;
				break;
			case 3:
				outerOffsetList[i].x = diamondLength - delta;
				break;
			default:
				break;
			}
			switch (outerTriangleUnit[i][0].y) {
			case 0:
				outerOffsetList[i].y = -2;
				break;
			case 1:
				outerOffsetList[i].y = diamondLength / 2 -
					delta - 1;
				break;
			case 2:
				outerOffsetList[i].y = diamondLength / 2;
				break;
			case 3:
				outerOffsetList[i].y = diamondLength- delta;
				break;
			default:
				break;
			}
		}
		for (int i = 0; i < Bevel.MAX_ORIENT; i++) {
			for (int j = 0; j <= 3; j++) {
				innerTriangleListX[i][j] = (j > 0) ?
					innerTriangleUnit[i][j].x * (diamondLength / 2 -
					delta) + innerTriangleListX[i][j - 1] :
					innerOffsetList[i].x;
				innerTriangleListY[i][j] = (j > 0) ?
					innerTriangleUnit[i][j].y * (diamondLength / 2 -
					delta) + innerTriangleListY[i][j - 1] :
					innerOffsetList[i].y;
				outerTriangleListX[i][j] = (j > 0) ?
					outerTriangleUnit[i][j].x * (diamondLength / 2 -
					delta) + outerTriangleListX[i][j - 1] :
					outerOffsetList[i].x;
				outerTriangleListY[i][j] = (j > 0) ?
					outerTriangleUnit[i][j].y * (diamondLength / 2 -
					delta) + outerTriangleListY[i][j - 1] :
					outerOffsetList[i].y;
			}
		}

		for (int i = 0; i < Bevel.MAX_CUBES; i++) {
			if (letterUnit[i].x == 0)
				letterList[i].x = diamondLength / 8 +
					letterOffset.x;
			else if (letterUnit[i].x == 1)
				letterList[i].x = 3 * diamondLength / 8 +
					letterOffset.x;
			else if (letterUnit[i].x == 2)
				letterList[i].x = 5 * diamondLength / 8  +
					letterOffset.x;
			else if (letterUnit[i].x == 3)
				letterList[i].x = 7 * diamondLength / 8 - 2 +
					letterOffset.x;
			if (letterUnit[i].y == 0)
				letterList[i].y = diamondLength / 8 + 2 +
					letterOffset.y;
			else if (letterUnit[i].y == 1)
				letterList[i].y = 5 * diamondLength / 16 + 2 +
					letterOffset.y;
			else if (letterUnit[i].y == 2)
				letterList[i].y = 13 * diamondLength / 20 + 2 +
					letterOffset.y;
			else if (letterUnit[i].y == 3)
				letterList[i].y = 7 * diamondLength / 8 - 3 +
					letterOffset.y;
		}
	}

	void drawFrame(boolean show) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			if (show) {
				Point[] coord = new Point[MAXXY + 1];
				Point letters;
				int fontSize = 2 * faceLength / 9;

				for (int i = 0; i <= MAXXY; i++) {
					coord[i] = new Point(
						i * viewLength + puzzleOffset.x,
						i * viewLength + puzzleOffset.y);
				}
				g.setFont(new Font("arial", 0, fontSize));
				g.setColor(bevel.currentForeground);
				g.drawLine(coord[1].x, coord[0].y, coord[2].x, coord[0].y);
				g.drawLine(coord[3].x, coord[1].y, coord[3].x, coord[2].y);
				g.drawLine(coord[1].x, coord[3].y, coord[2].x, coord[3].y);
				g.drawLine(coord[0].x, coord[1].y, coord[0].x, coord[2].y);
				letters = new Point(
					(coord[0].x + coord[1].x) / 2 - delta,
					(coord[0].y + coord[1].y) / 2);
				g.setColor(bevel.foreground);
				g.drawString("Front",
					letters.x - 5 * fontSize / 4,
					letters.y + 2 * fontSize / 5);
				letters.x = (coord[2].x + coord[3].x) / 2 - delta;
				letters.y = (coord[2].y + coord[3].y) / 2;
				g.drawString("Back",
					letters.x - fontSize,
					letters.y + fontSize / 5);
				g.setColor(bevel.currentForeground);
				if (vertical) {
					g.drawLine(coord[1].x, coord[0].y, coord[1].x, coord[4].y);
					g.drawLine(coord[2].x, coord[0].y, coord[2].x, coord[4].y);
					g.drawLine(coord[0].x, coord[1].y, coord[3].x, coord[1].y);
					g.drawLine(coord[0].x, coord[2].y, coord[3].x, coord[2].y);
					g.drawLine(coord[1].x, coord[4].y, coord[2].x, coord[4].y);
				} else {
					g.drawLine(coord[0].x, coord[1].y, coord[4].x, coord[1].y);
					g.drawLine(coord[0].x, coord[2].y, coord[4].x, coord[2].y);
					g.drawLine(coord[1].x, coord[0].y, coord[1].x, coord[3].y);
					g.drawLine(coord[2].x, coord[0].y, coord[2].x, coord[3].y);
					g.drawLine(coord[4].x, coord[1].y, coord[4].x, coord[2].y);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(bevel.background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void resizePuzzle() {
		coreSize = new Point(bevel.getSize().width,
			bevel.getSize().height);
		delta = 3;

		vertical = (coreSize.y >= coreSize.x);
		if (vertical)
			faceLength = java.lang.Math.min(coreSize.y / MAXY,
				coreSize.x / MAXX);
		else
			faceLength = java.lang.Math.min(coreSize.y / MAXX,
				coreSize.x / MAXY);
		faceLength = java.lang.Math.max(faceLength - delta - 1, 0);
		diamondLength = faceLength - delta + 1;
		viewLength = faceLength + delta;
		if (vertical) {
			puzzleSize = new Point(MAXX * (viewLength - 1) + delta,
				MAXY * (viewLength - 1) + delta);
		} else {
			puzzleSize = new Point(MAXY * (viewLength - 1) + delta,
				MAXX * (viewLength - 1) + delta);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		innerTriangleUnit[0][0] = new Point(2, 0);
		innerTriangleUnit[0][1] = new Point(1, 1);
		innerTriangleUnit[0][2] = new Point(-1, 0);
		innerTriangleUnit[0][3] = new Point(0, -1);
		innerTriangleUnit[1][0] = new Point(3, 2);
		innerTriangleUnit[1][1] = new Point(-1, 1);
		innerTriangleUnit[1][2] = new Point(0, -1);
		innerTriangleUnit[1][3] = new Point(1, 0);
		innerTriangleUnit[2][0] = new Point(1, 3);
		innerTriangleUnit[2][1] = new Point(-1, -1);
		innerTriangleUnit[2][2] = new Point(1, 0);
		innerTriangleUnit[2][3] = new Point(0, 1);
		innerTriangleUnit[3][0] = new Point(0, 1);
		innerTriangleUnit[3][1] = new Point(1, -1);
		innerTriangleUnit[3][2] = new Point(0, 1);
		innerTriangleUnit[3][3] = new Point(-1, 0);
		outerTriangleUnit[0][0] = new Point(2, 0);
		outerTriangleUnit[0][1] = new Point(1, 0);
		outerTriangleUnit[0][2] = new Point(0, 1);
		outerTriangleUnit[0][3] = new Point(-1, -1);
		outerTriangleUnit[1][0] = new Point(3, 2);
		outerTriangleUnit[1][1] = new Point(0, 1);
		outerTriangleUnit[1][2] = new Point(-1, 0);
		outerTriangleUnit[1][3] = new Point(1, -1);
		outerTriangleUnit[2][0] = new Point(1, 3);
		outerTriangleUnit[2][1] = new Point(-1, 0);
		outerTriangleUnit[2][2] = new Point(0, -1);
		outerTriangleUnit[2][3] = new Point(1, 1);
		outerTriangleUnit[3][0] = new Point(0, 1);
		outerTriangleUnit[3][1] = new Point(0, -1);
		outerTriangleUnit[3][2] = new Point(1, 0);
		outerTriangleUnit[3][3] = new Point(-1, 1);
		letterUnit[0] = new Point(3, 0);
		letterUnit[1] = new Point(3, 3);
		letterUnit[2] = new Point(0, 3);
		letterUnit[3] = new Point(0, 0);
		letterUnit[4] = new Point(2, 1);
		letterUnit[5] = new Point(2, 2);
		letterUnit[6] = new Point(1, 2);
		letterUnit[7] = new Point(1, 1);
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		bevel.drawAllPieces();
	}

	void selectPuzzle(int control, int x, int y) {
		if ((bevel.currentFace = positionToPiece(x, y)) != -1) { //sets currentPosition
			if (control != 0 || bevel.practiceChoice || !bevel.checkSolved()) {
				selectPaint = true;
				bevel.repaint();
				return;
			}
			bevel.currentFace = -1;
		}
		bevel.currentFace = -1;
	}

	void releasePuzzle(int control, int x, int y) {
		if (bevel.currentFace == -1)
			return;
		bevel.nextControl = control;
		int tempPosition = bevel.currentPosition;
		bevel.nextFace = positionToPiece(x, y); // sets currentPosition
		bevel.nextPosition = bevel.currentPosition;
		bevel.currentPosition = tempPosition;
		releasePaint = true;
		bevel.repaint();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			bevel.initializePuzzle();
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (bevel.currentFace != -1) {
				if (bevel.currentPosition == Bevel.MAX_ORIENT) {
					try {
						Thread.sleep(3 * bevel.delay);
					} catch (InterruptedException e) {
						// ignore
					}
					bevel.currentFace = -1;
					bevel.currentDirection = -1;
				} else if (bevel.currentPosition != -1)
					drawOuterTriangle(bevel.currentFace,
						bevel.currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (bevel.currentFace != -1) {
				drawOuterTriangle(bevel.currentFace,
					bevel.currentPosition, 0);
				bevel.selectPieces();
				bevel.currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			switch (bevel.currentControl) {
			case 2:
				bevel.moveAltCb(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection);
				break;
			case 1:
				bevel.moveControlCb(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection);
				break;
			default:
				bevel.movePieces(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection,
					bevel.currentControl);
				if (stackPaint && bevel.checkSolved()) {
					((BevelFrame) frame).callback(
						BevelInterface.ACTION_SOLVED,
						2);
				}
				break;
			}
			bevel.currentDirection = bevel.currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(bevel.delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			bevel.randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public void keyPressed(KeyEvent event) {
		int modifiers = event.getModifiersEx();

		switch (event.getKeyCode()) {
		case KeyEvent.VK_DIVIDE:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.CCW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_HOME:
		case KeyEvent.VK_NUMPAD7:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TOP,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_UP:
		case KeyEvent.VK_NUMPAD9:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.LEFT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.CW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.RIGHT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_END:
		case KeyEvent.VK_NUMPAD1:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BOTTOM,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_DOWN:
		case KeyEvent.VK_NUMPAD3:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		default:
		}
	}
}
