package com.sillycycle.bagleyd.bevel;

/*
 * @(#)Bevel3D.java
 *
 * Copyright 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;

/**
 * The <code>Bevel3D</code> class does all the 3D graphics for
 * <code>BevelFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Bevel3D extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int COORD3D = 12;
	static final int MAX_VIEWS = 2;

	static final double SQRT_3 = 1.732050808;

	static int multiply(int a) {
		return ((int) (a * SQRT_3 / 2.0));
	}

	static int divide(int a) {
		return ((int) (a * 2.0 / SQRT_3));
	}

	static final int[/*MAX_RECT*/] planeToCube =
		{6, 0, 6, 1, 2, 3, 6, 4, 6, 6, 5, 6};
	static int[/*MAX_FACES*/] cubeToPlane =
		{1, 3, 4, 5, 7, 10};

	Point pos = new Point(0, 0);
	Point puzzleSize, viewSize, cubeSize, faceSize, coreSize;
	Point puzzleOffset, letterOffset, viewMiddle;
	int faceDiagonal, cubeDiagonal, delta;
	int lastX, lastY;

	Point[][] faceLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT];
	Point[][] cubeLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT];
	Point[][] diamondLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT + 1];
	Point[][][] innerTriangleLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT][4];
	Point[][][] outerTriangleLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT][4];
	Point[][] letter3DList = new Point[Bevel.MAX_FACES][Bevel.MAX_CUBES];
	Point[][][] orientDiamond = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT][2];
	Point[][][] orientTriangle = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT][2];
	Point start = new Point(0, 0);

	JFrame      frame = null;
	Bevel       bevel = null;
	int         randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false, vertical = false;

	public Bevel3D(JFrame frame, Bevel bevel) {
		this.frame = frame;
		this.bevel = bevel;
		firstPaint = true;
	}

	public boolean checkStarted() {
		return bevel.checkStarted();
	}

	Point cubeOffset3D(int face, int corner) {
		Point point;

		point = diamondOffset3D(face);
		for (int side = 1; side <= corner; side++) {
			point.x += cubeLoc3D[face][side].x;
			point.y += cubeLoc3D[face][side].y;
		}
		return point;
	}

	static int mapTo3D(int face, int side) {
		switch (face) { // return corner
		case 0:
			return (side + 2) % Bevel.MAX_ORIENT;
		case 1:
		case 5:
			return side;
		case 2:
			return (side + 1) % Bevel.MAX_ORIENT;
		case 3:
		case 4:
			return (side + 3) % Bevel.MAX_ORIENT;
		default:
			System.err.println("mapTo3D: face " + face);
		}
		return -1;
	}

	static int mapFrom3D(int face, int corner) {
		switch (face) { // return side
		case 0:
			return (corner + 2) % Bevel.MAX_ORIENT;
		case 1:
		case 5:
			return corner;
		case 2:
			return (corner + 3) % Bevel.MAX_ORIENT;
		case 3:
		case 4:
			return (corner + 1) % Bevel.MAX_ORIENT;
		default:
			System.err.println("mapFrom3D: face " + face);
		}
		return -1;
	}

	Point diamondOffset3D(int face) {
		if (vertical) { // return starting pt
			switch (face) {
			case 0:
				return new Point(
					viewMiddle.x + delta - 1,
					viewMiddle.y - delta - 2);
			case 1:
				return new Point(
					viewMiddle.x - 2 * delta,
					viewMiddle.y);
			case 2:
				return new Point(
					viewMiddle.x + delta - 1,
					viewMiddle.y + 2 * delta - 1);
			case 3:
				return new Point(
					viewMiddle.x + 2 * delta,
					viewSize.y + viewMiddle.y +
					delta - 1);
			case 4:
				return new Point(
					viewMiddle.x - delta + 1,
					viewSize.y + viewMiddle.y -
					delta + 1);
			case 5:
				return new Point(
					viewMiddle.x - 2,
					viewSize.y + viewMiddle.y +
					2 * delta + 2);
			default:
				System.err.println("diamondOffset3D: face " +
					face);
			}
		} else {
			switch (face) {
			case 0:
				return new Point(
					viewMiddle.x,
					viewMiddle.y - 2 * delta + 1);
			case 1:
				return new Point(
					viewMiddle.x - delta - 2,
					viewMiddle.y + delta);
			case 2:
				return new Point(
					viewMiddle.x + 2 * delta - 1,
					viewMiddle.y + delta);
			case 3:
				return new Point(
					viewSize.x + viewMiddle.x -
					delta + 1,
					viewMiddle.y - delta);
			case 4:
				return new Point(
					viewSize.x + viewMiddle.x +
					delta - 1,
					viewMiddle.y + 2 * delta);
			case 5:
				return new Point(
					viewSize.x + viewMiddle.x +
					2 * delta + 2,
					viewMiddle.y - delta);
			default:
				System.err.println("diamondOffset3D: face " +
					face);
			}
		}
		return new Point(0, 0);
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}

	static void drawPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawInnerTriangle(int face, int position, int offsetPosition) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			int side;
			int rotation, l, letterX, letterY;
			Point startCoord, point;

			Color fillColor, lineColor;

			side = mapTo3D(face, position);
			startCoord = cubeOffset3D(face, side);
			point = diamondOffset3D(face);
			rotation = bevel.cubeLoc[face][position + Bevel.MAX_ORIENT].rotation;
			point.x += (cubeLoc3D[face][1].x  + cubeLoc3D[face][2].x) / 2;
			point.y += (cubeLoc3D[face][1].y  + cubeLoc3D[face][2].y) / 2;
			startCoord.x = (startCoord.x + point.x -
				(2 * innerTriangleLoc3D[face][side][1].x + innerTriangleLoc3D[face][side][2].x)) / 2;
			startCoord.y = (startCoord.y + point.y -
				(2 * innerTriangleLoc3D[face][side][1].y + innerTriangleLoc3D[face][side][2].y)) / 2;
			letterX = startCoord.x + letter3DList[face][side + Bevel.MAX_ORIENT].x;
			letterY = startCoord.y + letter3DList[face][side + Bevel.MAX_ORIENT].y;
			/* innerTriangleLoc3D[face][side][0].x = start.x;
			innertriangleLoc3D[face][side][0].y = start.y; */
			l = bevel.cubeLoc[face][position + Bevel.MAX_ORIENT].face;
			if (offsetPosition == 0) {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = bevel.faceColor[l];
					lineColor = bevel.borderColor;
				}
			} else {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = bevel.faceColor[l];
					fillColor = bevel.borderColor;
				}
			}
			int[] tempListX = new int[3];
			int[] tempListY = new int[3];

			for (int k = 0; k < 3; k++) {
				tempListX[k] =
					innerTriangleLoc3D[face][side][k].x +
					((k > 0) ?
					tempListX[k - 1] : 0);
				tempListY[k] =
					innerTriangleLoc3D[face][side][k].y +
					((k > 0) ?
					tempListY[k - 1] : 0);
			}
			fillPolygon(g, fillColor, lineColor,
				startCoord.x, startCoord.y,
				tempListX, tempListY, 3);
			if (bevel.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = cubeDiagonal / 5;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(bevel.faceNames.charAt(l));
				g.drawString(buf, letterX - 2, letterY + 2);
			}
			if (bevel.orientChoice) {
				g.setColor(lineColor);
				g.drawLine(letterX +
					orientTriangle[face][rotation][0].x,
					letterY +
					orientTriangle[face][rotation][0].y,
					letterX +
					orientTriangle[face][rotation][1].x,
					letterY +
					orientTriangle[face][rotation][1].y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawOuterTriangle(int face, int position, int offsetPosition) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			int side;
			int rotation, l, letterX, letterY;
			Point startCoord;
			Color fillColor, lineColor;

			side = mapTo3D(face, position);
			startCoord = cubeOffset3D(face, side);
			rotation = bevel.cubeLoc[face][position].rotation;
			letterX = startCoord.x + letter3DList[face][side].x;
			letterY = startCoord.y + letter3DList[face][side].y;
			/* outerTriangleLoc3D[face][side][0].x = start.x;
			outerTriangleLoc3D[face][side][0].y = start.y; */
			l = bevel.cubeLoc[face][position].face;
			if (offsetPosition == 0) {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = bevel.faceColor[l];
					lineColor = bevel.borderColor;
				}
			} else {
				if (bevel.mono) {
					if (bevel.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = bevel.faceColor[l];
					fillColor = bevel.borderColor;
				}
			}
			int[] tempListX = new int[3];
			int[] tempListY = new int[3];

			for (int k = 0; k < 3; k++) {
				tempListX[k] =
					outerTriangleLoc3D[face][side][k].x +
					((k > 0) ?
					tempListX[k - 1] : 0);
				tempListY[k] =
					outerTriangleLoc3D[face][side][k].y +
					((k > 0) ?
					tempListY[k - 1] : 0);
			}
			fillPolygon(g, fillColor, lineColor,
				startCoord.x, startCoord.y,
				tempListX, tempListY, 3);
			if (bevel.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = cubeDiagonal / 5;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(bevel.faceNames.charAt(l));
				g.drawString(buf, letterX - 2, letterY + 2);
			}
			if (bevel.orientChoice) {
				g.setColor(lineColor);
				g.drawLine(letterX +
					orientTriangle[face][rotation][0].x,
					letterY +
					orientTriangle[face][rotation][0].y,
					letterX +
					orientTriangle[face][rotation][1].x,
					letterY +
					orientTriangle[face][rotation][1].y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawFace(int face) {
		for (int position = 0; position < Bevel.MAX_ORIENT; position++) {
			drawInnerTriangle(face, position, 0);
			drawOuterTriangle(face, position, 0);
		}
	}

	void drawAllPieces() {
		for (int face = 0; face < Bevel.MAX_FACES; face++)
			drawFace(face);
	}

	void drawFrame(boolean show) {
		Graphics g = bevel.getGraphics();

		if (g == null)
			return;
		try {
			if (show) {
				Point coord = new Point(viewSize.x + puzzleOffset.x,
					viewSize.y + puzzleOffset.y);
				int fontSize = cubeDiagonal / 3;

				g.setFont(new Font("arial", 0, fontSize));
				g.setColor(bevel.foreground);
				if (vertical) {
					g.drawLine(0, coord.y,
					  coord.x + puzzleOffset.x + 1,
					  coord.y);
					g.drawString("Front", 2 * delta,
					  3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
					  -3 * delta - 10 * fontSize / 4 + coreSize.x,
					  -delta - 2 * fontSize / 5 + coreSize.y);
				} else {
					g.drawLine(coord.x, 0,
					  coord.x, coord.y + puzzleOffset.y + 1);
					g.drawString("Front", 2 * delta,
					  3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
					  -3 * delta - 10 * fontSize / 4 + coreSize.x,
					  -delta - 2 * fontSize / 5 + coreSize.y);
				}
				g.setColor(bevel.currentForeground);
				for (int face = 0; face < Bevel.MAX_FACES; face++) {
					int[] tempListX = new int[Bevel.MAX_ORIENT];
					int[] tempListY = new int[Bevel.MAX_ORIENT];

					for (int i = 0; i < Bevel.MAX_ORIENT; i++) {
						tempListX[i] =
							faceLoc3D[face][i].x +
							((i > 0) ?
							tempListX[i - 1] : 0);
						tempListY[i] =
							faceLoc3D[face][i].y +
							((i > 0) ?
							tempListY[i - 1] : 0);
					}
					drawPolygon(g, 0, 0,
						tempListX, tempListY,
						Bevel.MAX_ORIENT);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(bevel.background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void resizePieces() {
		int side, corner;
		Point[][] subcubeLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT];
		Point[][][] orientCubeLoc3D = new Point[2][Bevel.MAX_FACES][Bevel.MAX_ORIENT];
		Point[][] subdiamondLoc3D = new Point[Bevel.MAX_FACES][Bevel.MAX_ORIENT];
		Point[][][] orientDiamondLoc3D = new Point[2][Bevel.MAX_FACES][Bevel.MAX_ORIENT];
		/* letterOffset.x = -2;
		letterOffset.y = 4; */
		viewMiddle = new Point(faceSize.x + puzzleOffset.x,
			faceSize.y + puzzleOffset.y);
		for (int face = 0; face < Bevel.MAX_FACES; face++) {
			faceLoc3D[face][0] = new Point(viewMiddle.x,
				viewMiddle.y);
			for (int orient = 1; orient < Bevel.MAX_ORIENT; orient++) {
				faceLoc3D[face][orient] = new Point(faceSize.x,
					faceSize.y);
			}
		}
		if (vertical) {
			faceLoc3D[0][1].x /= -2;
			faceLoc3D[0][1].y /= -1;
			faceLoc3D[0][2].y = 0;
			faceLoc3D[0][3].x /= 2;

			faceLoc3D[1][1].x /= -2;
			faceLoc3D[1][2].x /= -2;
			faceLoc3D[1][2].y /= -1;
			faceLoc3D[1][3].x /= 2;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y = 0;
			faceLoc3D[2][2].x /= -2;
			faceLoc3D[2][3].x /= -1;
			faceLoc3D[2][3].y = 0;

			for (int face = Bevel.MAX_FACES / 2; face < Bevel.MAX_FACES; face++)
				faceLoc3D[face][0].y += viewSize.y + 3;

			faceLoc3D[3][1].x /= 2;
			faceLoc3D[3][1].y /= -1;
			faceLoc3D[3][2].x /= 2;
			faceLoc3D[3][3].x /= -2;

			faceLoc3D[4][1].x /= -1;
			faceLoc3D[4][1].y = 0;
			faceLoc3D[4][2].x /= 2;
			faceLoc3D[4][2].y /= -1;
			faceLoc3D[4][3].y = 0;

			faceLoc3D[5][1].x /= 2;
			faceLoc3D[5][2].x /= -1;
			faceLoc3D[5][2].y = 0;
			faceLoc3D[5][3].x /= -2;
			faceLoc3D[5][3].y /= -1;
		} else {
			faceLoc3D[0][1].x /= -1;
			faceLoc3D[0][1].y /= -2;
			faceLoc3D[0][2].y /= -2;
			faceLoc3D[0][3].y /= 2;

			faceLoc3D[1][1].x = 0;
			faceLoc3D[1][2].x /= -1;
			faceLoc3D[1][2].y /= -2;
			faceLoc3D[1][3].x = 0;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y /= -2;
			faceLoc3D[2][2].x = 0;
			faceLoc3D[2][3].y /= 2;
			faceLoc3D[2][3].x /= -1;

			for (int face = Bevel.MAX_FACES / 2; face < Bevel.MAX_FACES; face++)
				faceLoc3D[face][0].x += viewSize.x + 3;

			faceLoc3D[3][1].x /= -1;
			faceLoc3D[3][1].y /= 2;
			faceLoc3D[3][2].x = 0;
			faceLoc3D[3][2].y /= -1;
			faceLoc3D[3][3].y /= -2;

			faceLoc3D[4][1].y /= 2;
			faceLoc3D[4][2].x /= -1;
			faceLoc3D[4][2].y /= 2;
			faceLoc3D[4][3].x /= -1;
			faceLoc3D[4][3].y /= -2;

			faceLoc3D[5][1].x = 0;
			faceLoc3D[5][1].y /= -1;
			faceLoc3D[5][2].y /= 2;
			faceLoc3D[5][3].x = 0;
		}
		for (int face = 0; face < Bevel.MAX_FACES; face++) {
			int tempX, tempY;
			cubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			subcubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			orientCubeLoc3D[0][face][0] = new Point(0, 0);
			orientCubeLoc3D[1][face][0] = new Point(0, 0);
			for (int orient = 1; orient < Bevel.MAX_ORIENT; orient++) {
				cubeLoc3D[face][orient] = new Point(
					faceLoc3D[face][orient].x - 3 *
					delta * faceLoc3D[face][orient].x /
					faceSize.x,
					faceLoc3D[face][orient].y - 3 *
					delta * faceLoc3D[face][orient].y /
					faceSize.y);
				subcubeLoc3D[face][orient] = new Point(
					(faceLoc3D[face][orient].x - 5 *
					delta * faceLoc3D[face][orient].x /
					faceSize.x) / 2,
					(faceLoc3D[face][orient].y - 5 *
					delta * faceLoc3D[face][orient].y /
					faceSize.y) / 2);
				orientCubeLoc3D[0][face][orient] = new Point(
					(faceLoc3D[face][orient].x -
					5 * faceLoc3D[face][orient].x * delta /
					faceSize.x) / 4,
					(faceLoc3D[face][orient].y -
					5 * faceLoc3D[face][orient].y * delta /
					faceSize.y) / 4);
				orientCubeLoc3D[1][face][orient] = new Point(
					(faceLoc3D[face][orient].x -
					7 * faceLoc3D[face][orient].x * delta /
					faceSize.x) / 6,
					(faceLoc3D[face][orient].y -
					7 * faceLoc3D[face][orient].y * delta /
					faceSize.y) / 6);
			}
			tempX =  (subcubeLoc3D[face][1].x + subcubeLoc3D[face][3].x) / 2 -
				subcubeLoc3D[face][2].x;
			tempY =  (subcubeLoc3D[face][1].y + subcubeLoc3D[face][3].y) / 2 -
				subcubeLoc3D[face][2].y;
			for (int orient = 0; orient < Bevel.MAX_ORIENT; orient++)
				for (int j = 0; j < 4; j++) {
					innerTriangleLoc3D[face][orient][j] = new
						Point(0, 0);
					outerTriangleLoc3D[face][orient][j] = new
						Point(0, 0);
				}
			innerTriangleLoc3D[face][0][1].x = subcubeLoc3D[face][3].x;
			innerTriangleLoc3D[face][0][2].x = tempX - subcubeLoc3D[face][3].x;
			innerTriangleLoc3D[face][0][1].y = subcubeLoc3D[face][3].y;
			innerTriangleLoc3D[face][0][2].y = tempY - subcubeLoc3D[face][3].y;
			innerTriangleLoc3D[face][1][1].x = tempX;
			innerTriangleLoc3D[face][1][2].x = -subcubeLoc3D[face][3].x - tempX;
			innerTriangleLoc3D[face][1][1].y = tempY;
			innerTriangleLoc3D[face][1][2].y = -subcubeLoc3D[face][3].y - tempY;
			innerTriangleLoc3D[face][2][1].x = -subcubeLoc3D[face][3].x;
			innerTriangleLoc3D[face][2][2].x = subcubeLoc3D[face][3].x - tempX;
			innerTriangleLoc3D[face][2][1].y = -subcubeLoc3D[face][3].y;
			innerTriangleLoc3D[face][2][2].y = subcubeLoc3D[face][3].y - tempY;
			innerTriangleLoc3D[face][3][1].x = -tempX;
			innerTriangleLoc3D[face][3][2].x = subcubeLoc3D[face][3].x + tempX;
			innerTriangleLoc3D[face][3][1].y = -tempY;
			innerTriangleLoc3D[face][3][2].y = subcubeLoc3D[face][3].y + tempY;

			outerTriangleLoc3D[face][0][1].x = subcubeLoc3D[face][1].x;
			outerTriangleLoc3D[face][0][2].x = subcubeLoc3D[face][2].x -
				subcubeLoc3D[face][1].x;
			outerTriangleLoc3D[face][0][1].y = subcubeLoc3D[face][1].y;
			outerTriangleLoc3D[face][0][2].y = subcubeLoc3D[face][2].y -
				subcubeLoc3D[face][1].y;
			outerTriangleLoc3D[face][1][1].x = subcubeLoc3D[face][2].x;
			outerTriangleLoc3D[face][1][2].x = -subcubeLoc3D[face][1].x -
				subcubeLoc3D[face][2].x;
			outerTriangleLoc3D[face][1][1].y = subcubeLoc3D[face][2].y;
			outerTriangleLoc3D[face][1][2].y = -subcubeLoc3D[face][1].y -
				subcubeLoc3D[face][2].y;
			outerTriangleLoc3D[face][2][1].x = -subcubeLoc3D[face][1].x;
			outerTriangleLoc3D[face][2][2].x = subcubeLoc3D[face][1].x -
				subcubeLoc3D[face][2].x;
			outerTriangleLoc3D[face][2][1].y = -subcubeLoc3D[face][1].y;
			outerTriangleLoc3D[face][2][2].y = subcubeLoc3D[face][1].y -
				subcubeLoc3D[face][2].y;
			outerTriangleLoc3D[face][3][1].x = -subcubeLoc3D[face][2].x;
			outerTriangleLoc3D[face][3][2].x = subcubeLoc3D[face][1].x +
				subcubeLoc3D[face][2].x;
			outerTriangleLoc3D[face][3][1].y = -subcubeLoc3D[face][2].y;
			outerTriangleLoc3D[face][3][2].y = subcubeLoc3D[face][1].y +
				subcubeLoc3D[face][2].y;
			for (int orient = 0; orient < Bevel.MAX_ORIENT; orient++) {
				letter3DList[face][orient] = new Point(
					(2 * outerTriangleLoc3D[face][orient][1].x +
					outerTriangleLoc3D[face][orient][2].x) / 3,
					(2 * outerTriangleLoc3D[face][orient][1].y +
					outerTriangleLoc3D[face][orient][2].y) / 3);
				letter3DList[face][orient + Bevel.MAX_ORIENT] = new Point(
					(2 * innerTriangleLoc3D[face][orient][1].x +
					innerTriangleLoc3D[face][orient][2].x) / 3,
					(2 * innerTriangleLoc3D[face][orient][1].y +
					innerTriangleLoc3D[face][orient][2].y) / 3);
				innerTriangleLoc3D[face][orient][3].x =
					-innerTriangleLoc3D[face][orient][1].x - innerTriangleLoc3D[face][orient][2].x;
				innerTriangleLoc3D[face][orient][3].y =
					-innerTriangleLoc3D[face][orient][1].y - innerTriangleLoc3D[face][orient][2].y;
				outerTriangleLoc3D[face][orient][3].x =
					-outerTriangleLoc3D[face][orient][1].x - outerTriangleLoc3D[face][orient][2].x;
				outerTriangleLoc3D[face][orient][3].y =
					-outerTriangleLoc3D[face][orient][1].y - outerTriangleLoc3D[face][orient][2].y;
			}
		}
		cubeSize.x = faceSize.x - 2 * delta;
		cubeSize.y = faceSize.y - 2 * delta;
		cubeDiagonal = faceDiagonal - 2 * delta;
		//cubeDiag = faceDiagonal + 2 * delta;

		for (int face = 0; face < Bevel.MAX_FACES; face++) {
			int orient;

			for (orient = 0; orient < Bevel.MAX_ORIENT - 1; orient++) {
				diamondLoc3D[face][orient] = new Point(
					(cubeLoc3D[face][orient].x +
					cubeLoc3D[face][orient + 1].x) / 2,
					(cubeLoc3D[face][orient].y +
					cubeLoc3D[face][orient + 1].y) / 2);
				subdiamondLoc3D[face][orient] = new Point(
					(subcubeLoc3D[face][orient].x +
					subcubeLoc3D[face][orient + 1].x) / 2,
					(subcubeLoc3D[face][orient].y +
					subcubeLoc3D[face][orient + 1].y) / 2);
				orientDiamondLoc3D[0][face][orient] = new Point(
					(orientCubeLoc3D[0][face][orient].x +
					orientCubeLoc3D[0][face][orient + 1].x) / 2,
					(orientCubeLoc3D[0][face][orient].y +
					orientCubeLoc3D[0][face][orient + 1].y) / 2);
				orientDiamondLoc3D[1][face][orient] = new Point(
					(orientCubeLoc3D[1][face][orient].x +
					orientCubeLoc3D[1][face][orient + 1].x) / 2,
					(orientCubeLoc3D[1][face][orient].y +
					orientCubeLoc3D[1][face][orient + 1].y) / 2);
			}
			/* Its a parallelagram so take advantage of that */
			diamondLoc3D[face][orient] = new Point(
				(cubeLoc3D[face][Bevel.MAX_ORIENT - 1].x -
				cubeLoc3D[face][Bevel.MAX_ORIENT / 2].x) / 2,
				(cubeLoc3D[face][Bevel.MAX_ORIENT - 1].y -
				cubeLoc3D[face][Bevel.MAX_ORIENT / 2].y) / 2);
			diamondLoc3D[face][Bevel.MAX_ORIENT] = new Point(
				-diamondLoc3D[face][1].x -
				diamondLoc3D[face][2].x - diamondLoc3D[face][3].x,
				-diamondLoc3D[face][1].y -
				diamondLoc3D[face][2].y - diamondLoc3D[face][3].y);

			subdiamondLoc3D[face][orient] = new Point(
				(subcubeLoc3D[face][Bevel.MAX_ORIENT - 1].x -
				subcubeLoc3D[face][Bevel.MAX_ORIENT / 2].x) / 2,
				(subcubeLoc3D[face][Bevel.MAX_ORIENT - 1].y -
				subcubeLoc3D[face][Bevel.MAX_ORIENT / 2].y) / 2);
			orientDiamondLoc3D[0][face][orient] = new Point(
				(orientCubeLoc3D[0][face][Bevel.MAX_ORIENT - 1].x -
				orientCubeLoc3D[0][face][Bevel.MAX_ORIENT / 2].x) / 2,
				(orientCubeLoc3D[0][face][Bevel.MAX_ORIENT - 1].y -
				orientCubeLoc3D[0][face][Bevel.MAX_ORIENT / 2].y) / 2);
			orientDiamondLoc3D[1][face][orient] = new Point(
				(orientCubeLoc3D[1][face][Bevel.MAX_ORIENT - 1].x -
				orientCubeLoc3D[1][face][Bevel.MAX_ORIENT / 2].x) / 2,
				(orientCubeLoc3D[1][face][Bevel.MAX_ORIENT - 1].y -
				orientCubeLoc3D[1][face][Bevel.MAX_ORIENT / 2].y) / 2);

			corner = mapFrom3D(face, 1);
			orientDiamond[face][corner][0] = new Point(
				cubeLoc3D[face][1].x / 2,
				cubeLoc3D[face][1].y / 2);
			orientDiamond[face][corner][1] = new Point(
				orientDiamond[face][corner][0].x +
				(cubeLoc3D[face][2].x - subcubeLoc3D[face][2].x) / 2,
				orientDiamond[face][corner][0].y +
				(cubeLoc3D[face][2].y - subcubeLoc3D[face][2].y) / 2);
			orientTriangle[face][corner][0] = new Point(
				-orientCubeLoc3D[0][face][2].x / 2,
				-orientCubeLoc3D[0][face][2].y / 2);
			orientTriangle[face][corner][1] = new Point(
				-orientCubeLoc3D[1][face][2].x / 2,
				-orientCubeLoc3D[1][face][2].y / 2);
			for (orient = 1; orient < Bevel.MAX_ORIENT; orient++) {
				side = corner;
				corner = mapFrom3D(face, (orient + 1) % Bevel.MAX_ORIENT);
				orientDiamond[face][corner][0] = new Point(
					orientDiamond[face][side][0].x +
					diamondLoc3D[face][orient].x,
					orientDiamond[face][side][0].y +
					diamondLoc3D[face][orient].y);
				orientDiamond[face][corner][1] = new Point(
					orientDiamond[face][side][1].x +
					subdiamondLoc3D[face][orient].x,
					orientDiamond[face][side][1].y +
					subdiamondLoc3D[face][orient].y);
				orientTriangle[face][corner][0] = new Point(
					orientTriangle[face][side][0].x +
					orientDiamondLoc3D[0][face][orient].x,
					orientTriangle[face][side][0].y +
					orientDiamondLoc3D[0][face][orient].y);
				orientTriangle[face][corner][1] = new Point(
					orientTriangle[face][side][1].x +
					orientDiamondLoc3D[1][face][orient].x,
					orientTriangle[face][side][1].y +
					orientDiamondLoc3D[1][face][orient].y);
			}
		}
	}

	void narrowSelection() {
		//sets bevel.currentFace, bevel.currentPosition, bevel.currentDirection
		switch (bevel.currentDirection) {
		case Bevel.TOP:
		case Bevel.RIGHT:
		case Bevel.BOTTOM:
		case Bevel.LEFT:
			if (vertical) {
				if (bevel.currentFace == 1)
					bevel.currentDirection = (bevel.currentDirection + 3) % Bevel.MAX_ORIENT +
						2 * Bevel.MAX_ORIENT;
				else if (bevel.currentFace == 3)
					bevel.currentDirection = (bevel.currentDirection + 2) % Bevel.MAX_ORIENT +
						2 * Bevel.MAX_ORIENT;
			} else {
				switch (bevel.currentFace) {
				case 0:
					bevel.currentDirection = bevel.currentDirection +
						2 * Bevel.MAX_ORIENT;
					break;
				case 4:
					bevel.currentDirection = (bevel.currentDirection + 1) % Bevel.MAX_ORIENT +
						2 * Bevel.MAX_ORIENT;
					break;
				case 5:
					bevel.currentDirection = (bevel.currentDirection + 2) % Bevel.MAX_ORIENT;
					break;
				default:
					break;
				}
			}
			break;
		case Bevel.CCW:
		case Bevel.CW:
			break;
		case Bevel.TR:
		case Bevel.BR:
		case Bevel.BL:
		case Bevel.TL:
			if (vertical) {
				if (bevel.currentFace == 1)
					bevel.currentDirection = bevel.currentDirection % Bevel.MAX_ORIENT;
				else if (bevel.currentFace == 3)
					bevel.currentDirection = (bevel.currentDirection + 3) % Bevel.MAX_ORIENT;
			} else {
				switch (bevel.currentFace) {
				case 0:
					bevel.currentDirection = (bevel.currentDirection + 3) % Bevel.MAX_ORIENT;
					break;
				case 4:
					bevel.currentDirection = (bevel.currentDirection + 2) % Bevel.MAX_ORIENT;
					break;
				case 5:
					bevel.currentDirection = (bevel.currentDirection + 2) % Bevel.MAX_ORIENT + 2 * Bevel.MAX_ORIENT;
					break;
				default:
					break;
				}
			}
			break;
		default:
			bevel.currentFace = -1;
			return;
		}
	}

	int positionToPiece(int x, int y) { // sets bevel.currentPosition
		int u, v, found, face, i = 0, j = 0;
		int tempx = x, tempy = y;
		boolean front, tl, ur, ul;
		Point coord;

		if (vertical) {
			tempx -= viewMiddle.x;
			front = (tempy < viewSize.y + puzzleOffset.y);
			if (!front)
				tempy -= viewSize.y;
			tl = (tempy < viewMiddle.y);
			tempy -= viewMiddle.y;
			u = -faceSize.y * tempx + faceDiagonal * tempy;
			v = faceSize.y * tempx + faceDiagonal * tempy;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 0 : 1);
				else
					face = ((ul) ? 1 : 2);
			} else {
				if (tl)
					face = ((ul) ? 4 : 3);
				else
					face = ((ur) ? 3 : 5);
			}
		} else {
			tempy -= viewMiddle.y;
			front = (tempx < viewSize.x + puzzleOffset.x);
			if (!front)
				tempx -= (viewSize.x);
			tl = (tempx < viewMiddle.x);
			tempx -= viewMiddle.x;
			u = -faceSize.x * tempy + faceDiagonal * tempx;
			v = faceSize.x * tempy + faceDiagonal * tempx;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 1 : 0);
				else
					face = ((ul) ? 0 : 2);
			} else {
				if (tl)
					face = ((ul) ? 3 : 4);
				else
					face = ((ur) ? 4 : 5);
			}
		}
		/*found = 0;
		coord = diamondOffset3D(face);
		for (int side = 0; side < Bevel.MAX_ORIENT; side++) {
			int x1, y1, x2, y2;

			x1 = coord.x + orientDiamond[face][side][0].x;
			y1 = coord.y + orientDiamond[face][side][0].y;
			x2 = coord.x + orientDiamond[face][(side + 1) % Bevel.MAX_ORIENT][0].x;
			y2 = coord.y + orientDiamond[face][(side + 1) % Bevel.MAX_ORIENT][0].y;
			if ((x2 - x1) * (y - y1) <= (y2 - y1) * (x - x1)) {
				bevel.currentPosition = side;
				found++;
			}
		}
		if (found == 0)
			bevel.currentPosition = Bevel.MAX_ORIENT;
		else if (found > 1)
			return -1;
		return face;*/
		if (vertical)
			switch (face) {
			case 0:
				i = (tempx - 2 -
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSize.x / 2 + delta);
				j = (tempy + 2) / (cubeSize.y / 2 + delta) +
					1;
				break;
			case 1:
				i = (tempx + 4 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				j = (-tempx - 6 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonal / 2 + delta));
				break;
			case 2:
				i = (tempx - 4 +
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSize.x / 2 + delta);
				j = (tempy - 4) / (cubeSize.y / 2 + delta);
				break;
			case 3:
				i = (-tempx + 5 +
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				j = (-tempx + 7 -
					(tempy * faceDiagonal / faceSize.y)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				break;
			case 4:
				i = (tempx +
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSize.x / 2 + delta) + 1;
				j = (tempy - 2) / (cubeSize.y / 2 + delta) +
					1;
				break;
			case 5:
				i = (tempx + 2 -
					(tempy * faceDiagonal / faceSize.y)) /
					(cubeSize.x / 2 + delta) + 1;
				j = (tempy - 6) / (cubeSize.y / 2 + delta);
				break;
			default:
				return -1;
			}
		else
			switch (face) {
			case 0:
				i = (-tempy - 3 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonal / 2 + delta));
				j = (tempy + 1 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				break;
			case 1:
				i = (tempx + 2) / (cubeSize.x / 2 + delta) +
					1;
				j = (tempy - 3 -
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSize.y / 2 + delta);
				break;
			case 2:
				i = (tempx - 4) / (cubeSize.x / 2 + delta);
				j = (tempy - 6 +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSize.y / 2 + delta);
				break;
		       case 3:
				i = tempx / (cubeSize.x / 2 + delta) +
					1;
				j = (tempy +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSize.y / 2 + delta) + 1;
				break;
			case 4:
				i = (-tempy + 9 -
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				j = (-tempy + 7 +
					(tempx * faceDiagonal / faceSize.x)) /
					(2 * (cubeDiagonal / 2 + delta)) + 1;
				break;
			case 5:
				i = (-tempx + 7) / (cubeSize.x / 2 + delta) +
					1;
				j = (-tempy - 4 +
					(tempx * faceDiagonal / faceSize.x)) /
					(cubeSize.y / 2 + delta);
				break;
			default:
				return -1;
			}
		if (i < 0 || j < 0 || i >= 2 || j >= 2)
			return -1;
		if (i >= 1 && j < 1)
			bevel.currentPosition = 0;
		else if (i >= 1 && j >= 1)
			bevel.currentPosition = 1;
		else if (i < 1 && j >= 1)
			bevel.currentPosition = 2;
		else if (i < 1 && j < 1)
			bevel.currentPosition = 3;
		return face;
	}

	void resizePuzzle() {
		coreSize = new Point(bevel.getSize().width,
			bevel.getSize().height);
		delta = 4;
		vertical = (coreSize.y >= coreSize.x);
		if (vertical) {
			cubeSize = new Point(coreSize.x,
				coreSize.y / MAX_VIEWS);
			if (cubeSize.x >= divide(cubeSize.y)) {
				cubeSize.y = java.lang.Math.max((cubeSize.y -
					3 * delta) / 2 - delta - 2, 0);
				cubeSize.x = divide(cubeSize.y);
			} else {
				cubeSize.x = java.lang.Math.max((cubeSize.x -
					2 * delta - 7) / 2 - delta, 0);
				cubeSize.y = multiply(cubeSize.x);
			}
			faceSize = new Point(
				java.lang.Math.max(cubeSize.x + 2 * delta + 1, 1),
				java.lang.Math.max(cubeSize.y + 2 * delta + 1, 1));
			faceDiagonal = faceSize.x / 2;
			viewSize = new Point(2 * faceSize.x + 3,
				2 * faceSize.y + 3);
			puzzleSize = new Point(viewSize.x + 1,
				MAX_VIEWS * viewSize.y + 1);
		} else {
			cubeSize = new Point(coreSize.x / MAX_VIEWS,
				coreSize.y);
			if (cubeSize.y >= divide(cubeSize.x)) {
				cubeSize.x = java.lang.Math.max((cubeSize.x -
					3 * delta) / 2 - delta - 2, 0);
				cubeSize.y = divide(cubeSize.x);
			} else {
				cubeSize.y = java.lang.Math.max((cubeSize.y -
					2 * delta - 7) / 2 - delta, 0);
				cubeSize.x = multiply(cubeSize.y);
			}
			faceSize = new Point(
				java.lang.Math.max(cubeSize.x + 2 * delta + 1, 1),
				java.lang.Math.max(cubeSize.y + 2 * delta + 1, 1));
			faceDiagonal = faceSize.y / 2;
			viewSize = new Point(2 * faceSize.x + 3,
				2 * faceSize.y + 3);
			puzzleSize = new Point(MAX_VIEWS * viewSize.x + 1,
				viewSize.y + 1);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		bevel.drawAllPieces();
	}

	void selectPuzzle(int control, int x, int y) {
		if ((bevel.currentFace = positionToPiece(x, y)) != -1) { //sets currentPosition
			if (control != 0 || bevel.practiceChoice || !bevel.checkSolved()) {
				selectPaint = true;
				bevel.repaint();
				return;
			}
			bevel.currentFace = -1;
		}
		bevel.currentFace = -1;
	}

	void releasePuzzle(int control, int x, int y) {
		if (bevel.currentFace == -1)
			return;
		bevel.nextControl = control;
		int tempPosition = bevel.currentPosition;
		bevel.nextFace = positionToPiece(x, y); // sets currentPosition
		bevel.nextPosition = bevel.currentPosition;
		bevel.currentPosition = tempPosition;
		releasePaint = true;
		bevel.repaint();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			bevel.initializePuzzle();
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (bevel.currentFace != -1) {
				if (bevel.currentPosition == Bevel.MAX_ORIENT) {
					try {
						Thread.sleep((long) 3 * bevel.delay);
					} catch (InterruptedException e) {
						// ignore
					}
					bevel.currentFace = -1;
					bevel.currentDirection = -1;
				} else if (bevel.currentPosition != -1)
					drawOuterTriangle(bevel.currentFace,
						bevel.currentPosition, 1);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (bevel.currentFace != -1) {
				drawOuterTriangle(bevel.currentFace,
					bevel.currentPosition, 0);
				bevel.selectPieces();
				bevel.currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			switch (bevel.currentControl) {
			case 2:
				bevel.moveAltCb(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection);
				break;
			case 1:
				bevel.moveControlCb(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection);
				break;
			default:
				bevel.movePieces(bevel.currentFace,
					bevel.currentPosition,
					bevel.currentDirection,
					bevel.currentControl);
				if (stackPaint && bevel.checkSolved()) {
				((BevelFrame) frame).callback(
					BevelInterface.ACTION_SOLVED,
					3);
				}
				break;
			}
			bevel.currentDirection = bevel.currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(bevel.delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			bevel.randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public void keyPressed(KeyEvent event) {
		int modifiers = event.getModifiersEx();

		switch (event.getKeyCode()) {
		case KeyEvent.VK_F11:
		case KeyEvent.VK_DIVIDE:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.CCW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_HOME:
		case KeyEvent.VK_NUMPAD7:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TOP,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_UP:
		case KeyEvent.VK_NUMPAD9:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.TR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.LEFT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_F12:
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.CW,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.RIGHT,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_END:
		case KeyEvent.VK_NUMPAD1:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BL,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BOTTOM,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_DOWN:
		case KeyEvent.VK_NUMPAD3:
			bevel.movePuzzleInput(bevel.lastX, bevel.lastY,
				Bevel.BR,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		default:
		}
	}
}
