package com.sillycycle.bagleyd.bevel;

/*
 * @(#)BevelSolve.java
 *
 * Copyright 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>BevelSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class BevelSolve {
	private static JFrame frame;
	private static Bevel bevel;
	private static Random generator;
	private static final boolean DEBUG = false;
	//private static int counter; // debugging

	BevelSolve(JFrame frame, Bevel bevel) {
		BevelSolve.frame = frame;
		BevelSolve.bevel = bevel;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	boolean moveAPiece(int face, int position, int direction,
			boolean control) {
		if (!getFlag()) {
			return false;
		}
		bevel.movePuzzleDelay(face, position, direction,
			((control) ? 1 : 0), Bevel.DOUBLE);
		return true;
	}

	static final int MAX_CORNERS = 8;

	static final int bevelCorner[][]/*[MAX_CORNERS=8][3]*/ =
	{
		{1, 4, 5},
		{1, 2, 4},
		{0, 1, 5},
		{0, 2, 1},
		{3, 5, 4},
		{2, 3, 4},
		{0, 5, 3},
		{0, 3, 2}
	};

	static final int bevelCornerOrient[][]/*[MAX_CORNERS][3]*/ =
	{
		{2, 2, 3},
		{1, 2, 3},
		{3, 3, 2},
		{2, 3, 0},
		{1, 0, 1},
		{1, 2, 0},
		{0, 1, 0},
		{1, 3, 0}
	};

	/* Corner layout, bits correspond to position
	   2 6
	 2 3 7 6
	 0 1 5 4
	   0 4
	   2 6
	 */

	static int cornerColor(int corner, int n) {
		return bevel.cubeLoc[bevelCorner[corner][n]][bevelCornerOrient[corner][n]].face;
	}

	static int findOtherCorner(int corner, int color0, int color1) {
		int i, j, color, count;

		for (i = 0; i < MAX_CORNERS; i++) {
			if (i == corner)
				continue;
			count = 0;
			for (j = 0; j < 3; j++) {
				color = cornerColor(i, j);
				if (color == color0 || color == color1)
					count++;
				if (count == 2)
					return i;
			}
		}
		return MAX_CORNERS;
	}

	static int findCorner(int color0, int color1, int color2) {
		int i, j, color, count;

		for (i = 0; i < MAX_CORNERS; i++) {
			count = 0;
			for (j = 0; j < 3; j++) {
				color = cornerColor(i, j);
				if (color == color0 || color == color1 || color == color2)
					count++;
				if (count == 3)
					return i;
			}
		}
		return MAX_CORNERS;
	}

	void setSecondCorner() {
		int fixedCorner = 3;
		int color0 = cornerColor(fixedCorner, 0);
		int color1 = cornerColor(fixedCorner, 1);
		int foundCorner = findOtherCorner(fixedCorner, color0, color1);

		if (DEBUG) {	
			int targetCorner = 7;
			System.out.println("secondCorner " + targetCorner);
			System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 0:
			moveAPiece(4, 1, Bevel.LEFT, false);
			moveAPiece(4, 1, Bevel.TOP, false);
			moveAPiece(2, 1, Bevel.TOP, false);
			break;
		case 1:
			moveAPiece(2, 1, Bevel.LEFT, false);
			moveAPiece(2, 1, Bevel.TOP, false);
			break;
		case 2:
			moveAPiece(0, 0, Bevel.LEFT, false);
			moveAPiece(0, 1, Bevel.BOTTOM, false);
			break;
		case 4:
			moveAPiece(4, 1, Bevel.TOP, false);
			moveAPiece(2, 1, Bevel.TOP, false);
			break;
		case 5:
			moveAPiece(2, 1, Bevel.TOP, false);
			break;
		case 6:
			moveAPiece(0, 1, Bevel.BOTTOM, false);
			break;
		}
		/* rotate "turnstile" if needed */
		if (color0 != bevel.cubeLoc[0][1].face) {
			if (color0 == bevel.cubeLoc[2][0].face) {
				moveAPiece(3, 3, Bevel.RIGHT, false);
				moveAPiece(3, 0, Bevel.BOTTOM, false);
				moveAPiece(3, 1, Bevel.LEFT, false);
				moveAPiece(3, 2, Bevel.TOP, false);
			} else {
				moveAPiece(3, 3, Bevel.BOTTOM, false);
				moveAPiece(3, 2, Bevel.RIGHT, false);
				moveAPiece(3, 1, Bevel.TOP, false);
				moveAPiece(3, 0, Bevel.LEFT, false);
			}
		}
	}

	void setThirdCorner() {
		int fixedCorner = 3;
		int color0 = cornerColor(fixedCorner, 1);
		int color1 = cornerColor(fixedCorner, 2);
		int foundCorner = findOtherCorner(fixedCorner, color0, color1);

		if (DEBUG) {	
		int targetCorner = 1;
			System.out.println("thirdCorner " + targetCorner);
			System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 0:
			moveAPiece(4, 2, Bevel.TOP, false);
			break;
		case 2:
			moveAPiece(5, 2, Bevel.TOP, false);
			moveAPiece(4, 2, Bevel.TOP, false);
			break;
		case 4:
			moveAPiece(4, 1, Bevel.TOP, false);
			moveAPiece(2, 1, Bevel.LEFT, false);
			break;
		case 5:
			moveAPiece(2, 1, Bevel.LEFT, false);
			break;
		case 6:
			moveAPiece(5, 1, Bevel.TOP, false);
			moveAPiece(4, 1, Bevel.TOP, false);
			moveAPiece(2, 1, Bevel.LEFT, false);
			break;
		}
		/* rotate "turnstile" if needed */
		if (color0 != bevel.cubeLoc[2][2].face) {
			if (color0 == bevel.cubeLoc[1][1].face) {
				moveAPiece(4, 3, Bevel.RIGHT, false);
				moveAPiece(4, 0, Bevel.BOTTOM, false);
				moveAPiece(4, 1, Bevel.LEFT, false);
				moveAPiece(4, 2, Bevel.TOP, false);
			} else {
				moveAPiece(4, 3, Bevel.BOTTOM, false);
				moveAPiece(4, 2, Bevel.RIGHT, false);
				moveAPiece(4, 1, Bevel.TOP, false);
				moveAPiece(4, 0, Bevel.LEFT, false);
			}
		}
	}

	void setFourthCorner() {
		int fixedCorner = 3;
		int color0 = cornerColor(fixedCorner, 0);
		int color1 = cornerColor(fixedCorner, 2);
		int foundCorner = findOtherCorner(fixedCorner, color0, color1);

		if (DEBUG) {	
			int targetCorner = 2;
				System.out.println("forthCorner " + targetCorner);
				System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 0:
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			break;
		case 4:
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			break;
		case 5:
			moveAPiece(4, 0, Bevel.BOTTOM, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			break;
		case 6:
			moveAPiece(5, 1, Bevel.TOP, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			break;
		}
		/* rotate "turnstile" if needed */
		if (color0 != bevel.cubeLoc[0][3].face) {
			if (color0 == bevel.cubeLoc[1][3].face) {
				moveAPiece(5, 2, Bevel.RIGHT, false);
				moveAPiece(5, 1, Bevel.TOP, false);
				moveAPiece(5, 0, Bevel.LEFT, false);
				moveAPiece(5, 3, Bevel.BOTTOM, false);
			} else {
				moveAPiece(5, 2, Bevel.TOP, false);
				moveAPiece(5, 3, Bevel.RIGHT, false);
				moveAPiece(5, 0, Bevel.BOTTOM, false);
				moveAPiece(5, 1, Bevel.LEFT, false);
			}
		}
	}

	void turnOver() {
		moveAPiece(0, 0, Bevel.BOTTOM, true);
		moveAPiece(0, 0, Bevel.BOTTOM, true);
		moveAPiece(0, 0, Bevel.LEFT, true);
	}

	void setFifthCorner() {
		int color0 = bevel.cubeLoc[0][0].face;
		int color1 = bevel.cubeLoc[3][0].face;
		int color2 = bevel.cubeLoc[2][1].face;
		int foundCorner = findCorner(color0, color1, color2);

		if (DEBUG) {	
		int targetCorner = 7;
			System.out.println("fifthCorner " + targetCorner);
			System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 1:
			if (bevel.cubeLoc[1][1].face == color0) {
				moveAPiece(2, 2, Bevel.TOP, false);
				moveAPiece(0, 2, Bevel.RIGHT, false);
			} else if (bevel.cubeLoc[1][1].face == color1) {
				moveAPiece(2, 2, Bevel.TOP, false);
				moveAPiece(0, 2, Bevel.TOP, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
				moveAPiece(0, 0, Bevel.BOTTOM, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
			} else /* if (bevel.cubeLoc[1][1].face == color2) */ {
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
			}
			break;
		case 2:
			if (bevel.cubeLoc[5][2].face == color0) {
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(0, 2, Bevel.RIGHT, false);
			} else if (bevel.cubeLoc[5][2].face == color2) {
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
			} else /* if (bevel.cubeLoc[5][2].face == color1) */ {
				moveAPiece(0, 3, Bevel.RIGHT, false);
				moveAPiece(0, 0, Bevel.BOTTOM, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
			}
			break;
		case 3:
			if (bevel.cubeLoc[2][3].face == color0) {
				moveAPiece(0, 2, Bevel.LEFT, false);
			} else if (bevel.cubeLoc[2][3].face == color1) {
				moveAPiece(0, 2, Bevel.TOP, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
				moveAPiece(0, 0, Bevel.BOTTOM, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
			} else /* if (bevel.cubeLoc[2][3].face == color2) */ {
				moveAPiece(2, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 2, Bevel.LEFT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 2, Bevel.LEFT, false);
			}
			break;
		case 7:
			if (bevel.cubeLoc[2][0].face == color0) {
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 1, Bevel.LEFT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 2, Bevel.TOP, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
			} else if (bevel.cubeLoc[2][0].face == color1) {
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 0, Bevel.LEFT, false);
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(0, 2, Bevel.RIGHT, false);
			}
			break;
		}
	}

	void setSixthCorner() {
		int color0 = bevel.cubeLoc[2][1].face;
		int color1 = bevel.cubeLoc[4][0].face;
		int color2 = bevel.cubeLoc[1][2].face;
		int foundCorner = findCorner(color0, color1, color2);

		if (DEBUG) {	
			int targetCorner = 1;
			System.out.println("sixthCorner " + targetCorner);
			System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 1:
			if (bevel.cubeLoc[1][1].face == color0) {
				moveAPiece(2, 2, Bevel.TOP, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 1, Bevel.LEFT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
			} else if (bevel.cubeLoc[1][1].face == color1) {
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 0, Bevel.LEFT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
			}
			break;
		case 2:
			if (bevel.cubeLoc[0][3].face == color0) {
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
			} else if (bevel.cubeLoc[0][3].face == color1) {
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 1, Bevel.LEFT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
			} else /* if (bevel.cubeLoc[0][3].face == color2) */ {
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 0, Bevel.LEFT, false);
				moveAPiece(2, 1, Bevel.TOP, false);
				moveAPiece(2, 2, Bevel.RIGHT, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
			}
			break;
		case 3:
			if (bevel.cubeLoc[1][0].face == color0) {
				moveAPiece(2, 3, Bevel.BOTTOM, false);
			} else if (bevel.cubeLoc[1][0].face == color1) {
				moveAPiece(2, 3, Bevel.RIGHT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 1, Bevel.LEFT, false);
				moveAPiece(2, 0, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.RIGHT, false);
			} else /* if (bevel.cubeLoc[1][0].face == color2) */ {
				moveAPiece(0, 2, Bevel.RIGHT, false);
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 0, Bevel.LEFT, false);
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 2, Bevel.RIGHT, false);
				moveAPiece(0, 3, Bevel.BOTTOM, false);
				moveAPiece(2, 3, Bevel.BOTTOM, false);
			}
			break;
		}
	}

	void setLastCorners() {
		int color0 = bevel.cubeLoc[0][0].face;
		int color1 = bevel.cubeLoc[1][2].face;
		int color2 = bevel.cubeLoc[5][1].face;
		int foundCorner = findCorner(color0, color1, color2);

		if (DEBUG) {	
			int targetCorner = 2;
			System.out.println("seventhCorner " + targetCorner);
			System.out.println("foundCorner " + foundCorner);
		}
		switch (foundCorner) {
		case 2:
			if (bevel.cubeLoc[1][3].face == color0) {
				moveAPiece(1, 3, Bevel.BOTTOM, false);
				moveAPiece(1, 0, Bevel.BOTTOM, false);
				moveAPiece(1, 1, Bevel.LEFT, false);
				moveAPiece(1, 2, Bevel.TOP, false);
				moveAPiece(1, 1, Bevel.TOP, false);
				moveAPiece(1, 0, Bevel.LEFT, false);
			} else if (bevel.cubeLoc[1][3].face == color2) {
				moveAPiece(0, 2, Bevel.RIGHT, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 1, Bevel.LEFT, false);
				moveAPiece(0, 0, Bevel.LEFT, false);
				moveAPiece(0, 3, Bevel.BOTTOM, false);
			}
			break;
		case 3:
			if (bevel.cubeLoc[1][0].face == color0) {
				moveAPiece(0, 2, Bevel.TOP, false);
			} else if (bevel.cubeLoc[1][0].face == color1) {
				moveAPiece(1, 3, Bevel.BOTTOM, false);
				moveAPiece(1, 0, Bevel.BOTTOM, false);
				moveAPiece(1, 1, Bevel.LEFT, false);
				moveAPiece(1, 2, Bevel.TOP, false);
				moveAPiece(1, 1, Bevel.TOP, false);
			} else /* if (bevel.cubeLoc[1][0].face == color2) */ {
				moveAPiece(0, 2, Bevel.RIGHT, false);
				moveAPiece(0, 3, Bevel.RIGHT, false);
				moveAPiece(0, 1, Bevel.TOP, false);
				moveAPiece(0, 1, Bevel.LEFT, false);
				moveAPiece(0, 0, Bevel.LEFT, false);
			}
			break;
		}
	}


	/* 6 moves */
	void swap4(int corner, int face) {
		int i;
		int which = (generator.nextInt(2) != 0) ? 1 : 0;

		switch (corner) {
		case 0:
			switch (face) {
			case 1:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(1, 1, Bevel.LEFT, false);
					else
						moveAPiece(1, 3, Bevel.BOTTOM, false);
				}
				break;
			case 4:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(4, 1, Bevel.LEFT, false);
					else
						moveAPiece(4, 3, Bevel.BOTTOM, false);
				}
				break;
			case 5:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(5, 0, Bevel.LEFT, false);
					else
						moveAPiece(5, 2, Bevel.TOP, false);
				}
				break;
			}
			break;
		case 1:
			switch (face) {
			case 1:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(1, 0, Bevel.BOTTOM, false);
					else
						moveAPiece(1, 2, Bevel.LEFT, false);
				}
				break;
			case 2:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(2, 1, Bevel.LEFT, false);
					else
						moveAPiece(2, 3, Bevel.BOTTOM, false);
				}
				break;
			case 4:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(4, 0, Bevel.LEFT, false);
					else
						moveAPiece(4, 2, Bevel.TOP, false);
				}
				break;
			}
			break;
		case 2:
			switch (face) {
			case 0:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(0, 0, Bevel.LEFT, false);
					else
						moveAPiece(0, 2, Bevel.TOP, false);
				}
				break;
			case 1:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(1, 0, Bevel.LEFT, false);
					else
						moveAPiece(1, 2, Bevel.TOP, false);
				}
				break;
			case 5:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(5, 1, Bevel.LEFT, false);
					else
						moveAPiece(5, 3, Bevel.BOTTOM, false);
				}
				break;
			}
			break;
		case 3:
			switch (face) {
			case 0:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(0, 1, Bevel.LEFT, false);
					else
						moveAPiece(0, 3, Bevel.BOTTOM, false);
				}
				break;
			case 1:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(1, 1, Bevel.TOP, false);
					else
						moveAPiece(1, 3, Bevel.RIGHT, false);
				}
				break;
			case 2:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(2, 0, Bevel.LEFT, false);
					else
						moveAPiece(2, 2, Bevel.TOP, false);
				}
				break;
			}
			break;
		case 4:
			switch (face) {
			case 3:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(3, 0, Bevel.BOTTOM, false);
					else
						moveAPiece(3, 2, Bevel.RIGHT, false);
				}
				break;
			case 4:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(4, 0, Bevel.BOTTOM, false);
					else
						moveAPiece(4, 2, Bevel.RIGHT, false);
				}
				break;
			case 5:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(5, 1, Bevel.TOP, false);
					else
						moveAPiece(5, 3, Bevel.RIGHT, false);
				}
				break;
			}
			break;
		case 5:
			switch (face) {
			case 2:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(2, 0, Bevel.BOTTOM, false);
					else
						moveAPiece(2, 2, Bevel.RIGHT, false);
				}
				break;
			case 3:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(3, 0, Bevel.RIGHT, false);
					else
						moveAPiece(3, 3, Bevel.BOTTOM, false);
				}
				break;
			case 4:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(4, 1, Bevel.TOP, false);
					else
						moveAPiece(4, 3, Bevel.RIGHT, false);
				}
				break;
			}
			break;
		case 6:
			switch (face) {
			case 0:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(0, 1, Bevel.TOP, false);
					else
						moveAPiece(0, 3, Bevel.RIGHT, false);
				}
				break;
			case 3:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(3, 1, Bevel.TOP, false);
					else
						moveAPiece(3, 3, Bevel.RIGHT, false);
				}
				break;
			case 5:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(5, 0, Bevel.BOTTOM, false);
					else
						moveAPiece(5, 2, Bevel.RIGHT, false);
				}
				break;
			}
			break;
		case 7:
			switch (face) {
			case 0:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(0, 1, Bevel.BOTTOM, false);
					else
						moveAPiece(0, 2, Bevel.RIGHT, false);
				}
				break;
			case 2:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(2, 1, Bevel.TOP, false);
					else
						moveAPiece(2, 3, Bevel.RIGHT, false);
				}
				break;
			case 3:
				for (i = 0; i < 6; i++) {
					if ((i + which) % 2 == 0)
						moveAPiece(3, 0, Bevel.LEFT, false);
					else
						moveAPiece(3, 2, Bevel.TOP, false);
				}
				break;
			}
			break;
		}
	}

	/* 22 moves */
	void swap3AroundCorner(int corner, int side, boolean cw) {
		int i;

		switch (corner) {
		case 0:
			switch (side) {
			case 1:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 0, Bevel.LEFT, false);
						moveAPiece(1, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 1, Bevel.LEFT, false);
						moveAPiece(1, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(5, 2, Bevel.TOP, false);
					}
				}
				break;
			case 4:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 3, Bevel.BOTTOM, false);
						moveAPiece(4, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(5, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 2, Bevel.TOP, false);
						moveAPiece(4, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 1, Bevel.LEFT, false);
					}
				}
				break;
			case 5:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 3, Bevel.BOTTOM, false);
						moveAPiece(5, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 1, Bevel.LEFT, false);
						moveAPiece(5, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(4, 1, Bevel.LEFT, false);
					}
				}
				break;
			}
			break;
		case 1:
			switch (side) {
			case 1:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 0, Bevel.LEFT, false);
						moveAPiece(1, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(2, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 1, Bevel.LEFT, false);
						moveAPiece(1, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 2, Bevel.TOP, false);
					}
				}
				break;
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 2, Bevel.RIGHT, false);
						moveAPiece(2, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 2, Bevel.TOP, false);
						moveAPiece(2, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			case 4:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 3, Bevel.BOTTOM, false);
						moveAPiece(4, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 0, Bevel.BOTTOM, false);
						moveAPiece(4, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(2, 1, Bevel.LEFT, false);
					}
				}
				break;
			}
			break;
		case 2:
			switch (side) {
			case 0:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 0, Bevel.LEFT, false);
						moveAPiece(5, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(0, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 2, Bevel.TOP, false);
						moveAPiece(5, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 2, Bevel.TOP, false);
					}
				}
				break;
			case 1:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 0, Bevel.LEFT, false);
						moveAPiece(1, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(5, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 1, Bevel.LEFT, false);
						moveAPiece(1, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 2, Bevel.TOP, false);
					}
				}
				break;
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 3, Bevel.BOTTOM, false);
						moveAPiece(0, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 2, Bevel.TOP, false);
						moveAPiece(0, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(5, 1, Bevel.LEFT, false);
					}
				}
				break;
			}
			break;
		case 3:
			switch (side) {
			case 0:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 1, Bevel.TOP, false);
						moveAPiece(0, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(2, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 2, Bevel.TOP, false);
						moveAPiece(0, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 1:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 0, Bevel.LEFT, false);
						moveAPiece(1, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 1, Bevel.LEFT, false);
						moveAPiece(1, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(2, 2, Bevel.TOP, false);
					}
				}
				break;
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 3, Bevel.BOTTOM, false);
						moveAPiece(2, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(1, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(1, 3, Bevel.RIGHT, false);
						moveAPiece(2, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 1, Bevel.LEFT, false);
					}
				}
				break;
			}
			break;
		case 4:
			switch (side) {
			case 3:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 2, Bevel.RIGHT, false);
						moveAPiece(3, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(5, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 3, Bevel.RIGHT, false);
						moveAPiece(3, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			case 4:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 1, Bevel.TOP, false);
						moveAPiece(4, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 0, Bevel.BOTTOM, false);
						moveAPiece(4, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(5, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 5:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 2, Bevel.RIGHT, false);
						moveAPiece(5, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(4, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 0, Bevel.BOTTOM, false);
						moveAPiece(5, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			}
			break;
		case 5:
			switch (side) {
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 1, Bevel.TOP, false);
						moveAPiece(2, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 3, Bevel.BOTTOM, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 1, Bevel.LEFT, false);
						moveAPiece(2, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 3:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 2, Bevel.RIGHT, false);
						moveAPiece(3, 3, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(4, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(4, 3, Bevel.RIGHT, false);
						moveAPiece(3, 1, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(2, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			case 4:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 3, Bevel.BOTTOM, false);
						moveAPiece(4, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(2, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 0, Bevel.BOTTOM, false);
						moveAPiece(4, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 1, Bevel.LEFT, false);
					}
				}
				break;
			}
			break;
		case 6:
			switch (side) {
			case 0:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 1, Bevel.TOP, false);
						moveAPiece(5, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 3, Bevel.RIGHT, false);
						moveAPiece(5, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(0, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 1, Bevel.TOP, false);
						moveAPiece(0, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(5, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 0, Bevel.BOTTOM, false);
						moveAPiece(0, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 3:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(5, 2, Bevel.RIGHT, false);
						moveAPiece(3, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 3, Bevel.RIGHT, false);
						moveAPiece(3, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(5, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			}
			break;
		case 7:
			switch (side) {
			case 0:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 1, Bevel.TOP, false);
						moveAPiece(0, 2, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 0, Bevel.LEFT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 2, Bevel.TOP, false);
						moveAPiece(0, 0, Bevel.BOTTOM, false);
						if (i % 4 != 3)
							moveAPiece(2, 3, Bevel.RIGHT, false);
					}
				}
				break;
			case 2:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(3, 0, Bevel.LEFT, false);
						moveAPiece(2, 1, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 2, Bevel.RIGHT, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 0, Bevel.BOTTOM, false);
						moveAPiece(2, 3, Bevel.RIGHT, false);
						if (i % 4 != 3)
							moveAPiece(3, 2, Bevel.TOP, false);
					}
				}
				break;
			case 3:
				if (cw) {
					for (i = 0; i < 8; i++) {
						moveAPiece(0, 2, Bevel.RIGHT, false);
						moveAPiece(3, 0, Bevel.LEFT, false);
						if (i % 4 != 3)
							moveAPiece(2, 1, Bevel.TOP, false);
					}
				} else {
					for (i = 0; i < 8; i++) {
						moveAPiece(2, 3, Bevel.RIGHT, false);
						moveAPiece(3, 2, Bevel.TOP, false);
						if (i % 4 != 3)
							moveAPiece(0, 0, Bevel.BOTTOM, false);
					}
				}
				break;
			}
			break;
		}
	}

	void setFace0Pos4(int color) {
		int face = 0;
		int pos = 4;
		boolean which = (generator.nextInt(2) != 0);

		if (bevel.cubeLoc[face][pos].face == color) {
			return;
		} else if (bevel.cubeLoc[3][7].face == color) {
			swap4(6, 3);
		} else if (bevel.cubeLoc[5][6].face == color) {
			swap4(6, 5);
		} else if (bevel.cubeLoc[1][6].face == color) {
			moveAPiece(0, 0, Bevel.LEFT, false);
			swap4(2, 1);
			moveAPiece(0, 0, Bevel.LEFT, false);
		} else if (bevel.cubeLoc[2][5].face == color) {
			moveAPiece(0, 0, Bevel.BOTTOM, false);
			swap4(7, 2);
			moveAPiece(0, 0, Bevel.BOTTOM, false);
		} else if (bevel.cubeLoc[4][7].face == color) {
			if (which) {
				moveAPiece(0, 0, Bevel.BOTTOM, false);
				moveAPiece(3, 3, Bevel.BOTTOM, false);
				swap4(5, 4);
				moveAPiece(3, 3, Bevel.BOTTOM, false);
				moveAPiece(0, 0, Bevel.BOTTOM, false);
			} else {
				moveAPiece(0, 0, Bevel.LEFT, false);
				moveAPiece(5, 2, Bevel.TOP, false);
				swap4(0, 4);
				moveAPiece(5, 2, Bevel.TOP, false);
				moveAPiece(0, 0, Bevel.LEFT, false);
			}
		} else {
				System.out.println("Fix face " + face + " pos " + pos);
		}
	}

	void setFace2Pos7(int color) {
		int face = 2;
		int pos = 7;

		if (bevel.cubeLoc[face][pos].face == color) {
			return;
		} else if (bevel.cubeLoc[1][5].face == color) {
			swap4(1, 2);
		} else if (bevel.cubeLoc[4][6].face == color) {
			moveAPiece(2, 3, Bevel.BOTTOM, false);
			swap4(1, 4);
			moveAPiece(2, 3, Bevel.BOTTOM, false);
		} else if (bevel.cubeLoc[5][4].face == color) {
			moveAPiece(2, 3, Bevel.BOTTOM, false);
			moveAPiece(1, 1, Bevel.LEFT, false);
			swap4(4, 4);
			moveAPiece(1, 1, Bevel.LEFT, false);
			moveAPiece(2, 3, Bevel.BOTTOM, false);
		} else if (bevel.cubeLoc[3][4].face == color) {
			moveAPiece(2, 3, Bevel.BOTTOM, false);
			moveAPiece(1, 1, Bevel.LEFT, false);
			moveAPiece(4, 2, Bevel.RIGHT, false);
			swap4(4, 3);
			moveAPiece(4, 2, Bevel.RIGHT, false);
			moveAPiece(1, 1, Bevel.LEFT, false);
			moveAPiece(2, 3, Bevel.BOTTOM, false);
		} else {
				System.out.println("Fix face " + face + " pos " + pos);
		}
	}

	void setFace1Pos4(int color) {
		int face = 1;
		int pos = 4;

		if (bevel.cubeLoc[face][pos].face == color) {
			return;
		} else if (bevel.cubeLoc[2][6].face == color) {
			swap4(1, 1);
		} else if (bevel.cubeLoc[4][4].face == color) {
			/* collides with its opposite */
			/*moveAPiece(1, 0, Bevel.BOTTOM, false);
			swap4(1, 4);
			moveAPiece(1, 0, Bevel.BOTTOM, false);*/
			swap3AroundCorner(1, 4, true);
		} else if (bevel.cubeLoc[3][5].face == color) {
			moveAPiece(1, 0, Bevel.BOTTOM, false);
			moveAPiece(2, 2, Bevel.RIGHT, false);
			swap4(4, 4);
			moveAPiece(2, 2, Bevel.RIGHT, false);
			moveAPiece(1, 0, Bevel.BOTTOM, false);
		} else if (bevel.cubeLoc[5][5].face == color) {
			moveAPiece(1, 0, Bevel.BOTTOM, false);
			moveAPiece(2, 2, Bevel.RIGHT, false);
			moveAPiece(4, 0, Bevel.BOTTOM, false);
			swap4(4, 5);
			moveAPiece(4, 0, Bevel.BOTTOM, false);
			moveAPiece(2, 2, Bevel.RIGHT, false);
			moveAPiece(1, 0, Bevel.BOTTOM, false);
		} else {
				System.out.println("Fix face " + face + " pos " + pos);
		}
	}

	void setFace0() {
		int face = 0;
		int pos = 0;
		boolean which = (generator.nextInt(2) != 0);
		int i;
		int color = bevel.cubeLoc[face][pos].face;

		if (DEBUG) {	
			System.out.println("face " + face + ": zeroethEdge " + color);
		}
		setFace0Pos4(color);
		for (i = 0; i < Bevel.MAX_ORIENT - 1; i++) {
			moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
			color = bevel.cubeLoc[face][pos].face;
			setFace0Pos4(color);
		}
	}

	void setFaceTop() {
		int face = 2;
		int pos = 0;
		boolean which = (generator.nextInt(2) != 0);
		int i;
		int color = bevel.cubeLoc[face][pos].face;

		if (DEBUG) {	
			System.out.println("face " + face + ": zeroethEdge " + color);
		}
		setFace2Pos7(color);
		for (i = 0; i < Bevel.MAX_ORIENT - 1; i++) {
			moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
			color = bevel.cubeLoc[face][pos].face;
			setFace2Pos7(color);
		}
		face = 1;
		which = (generator.nextInt(2) != 0);
		color = bevel.cubeLoc[face][pos].face;
		setFace1Pos4(color);
		for (i = 0; i < Bevel.MAX_ORIENT - 1; i++) {
			moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
			color = bevel.cubeLoc[face][pos].face;
			setFace1Pos4(color);
		}
		moveAPiece(0, 0, Bevel.BOTTOM, true);
		moveAPiece(0, 0, Bevel.BOTTOM, true);
	}

	void fixATriple() {
		int corner = 3;
		int i, count = 0, pos = 0;
		int[] faceColor = new int[3];
		int[] facetColor = new int[3];

		for (i = 0; i < 3; i++) {
			switch (i) {
			case 0:
				pos = 6;
				break;
			case 1:
				pos = 4;
				break;
			case 2:
				pos = 7;
				break;
			}
			facetColor[i] = bevel.cubeLoc[bevelCorner[corner][i]][pos].face;
			faceColor[i] = bevel.cubeLoc[bevelCorner[corner][i]][0].face;
			if (facetColor[i] != faceColor[i])
				count++;
		}
		if (count == 3) {
			if (facetColor[2] == faceColor[1]) {
				swap3AroundCorner(3, 1, false);
			} else /* if (facetColor[1] == faceColor[2]) */ {
				swap3AroundCorner(3, 2, true);
			}
		}
	}

	void fixTriples() {
		int i;
		boolean which = (generator.nextInt(2) != 0);

		fixATriple();
		for (i = 0; i < Bevel.MAX_ORIENT - 1; i++) {
			moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
			fixATriple();
		}
	}

	boolean checkOddSplitDouble() {
		int count = 0;

		if (bevel.cubeLoc[0][5].face != bevel.cubeLoc[0][0].face)
			count++;
		if (bevel.cubeLoc[0][7].face != bevel.cubeLoc[0][0].face)
			count++;
		if (bevel.cubeLoc[1][4].face != bevel.cubeLoc[1][0].face)
			count++;
		if (bevel.cubeLoc[2][7].face != bevel.cubeLoc[2][0].face)
			count++;
		if (bevel.cubeLoc[3][4].face != bevel.cubeLoc[3][0].face)
			count++;
		if (bevel.cubeLoc[5][5].face != bevel.cubeLoc[5][0].face)
			count++;
		if (count == 4)
			return true;
		return false;
	}

	boolean checkEvenSplitDouble() {
		int count = 0;

		if (bevel.cubeLoc[0][4].face != bevel.cubeLoc[0][0].face)
			count++;
		if (bevel.cubeLoc[0][6].face != bevel.cubeLoc[0][0].face)
			count++;
		if (bevel.cubeLoc[1][7].face != bevel.cubeLoc[1][0].face)
			count++;
		if (bevel.cubeLoc[2][4].face != bevel.cubeLoc[2][0].face)
			count++;
		if (bevel.cubeLoc[3][7].face != bevel.cubeLoc[3][0].face)
			count++;
		if (bevel.cubeLoc[5][6].face != bevel.cubeLoc[5][0].face)
			count++;
		if (count == 4)
			return true;
		return false;
	}

	boolean checkConnectedDouble() {
		int i, count;
		boolean which = (generator.nextInt(2) != 0);

		for (i = 0; i < Bevel.MAX_ORIENT; i++) {
			count = 0;
			if (bevel.cubeLoc[0][5].face != bevel.cubeLoc[0][0].face)
				count++;
			if (bevel.cubeLoc[0][6].face != bevel.cubeLoc[0][0].face)
				count++;
			if (bevel.cubeLoc[1][7].face != bevel.cubeLoc[1][0].face)
				count++;
			if (bevel.cubeLoc[2][4].face != bevel.cubeLoc[2][0].face)
				count++;
			if (bevel.cubeLoc[2][7].face != bevel.cubeLoc[2][0].face)
				count++;
			if (bevel.cubeLoc[3][4].face != bevel.cubeLoc[3][0].face)
				count++;
			if (count == 4)
				return true; /* can only be one left */
			moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
		}
		return false;
	}

	void fixOddSplitDouble() {
		boolean doubleRight = false, right = false;
		boolean doubleLeft = false, left = false;

		if (bevel.cubeLoc[2][7].face == bevel.cubeLoc[2][1].face) {
			moveAPiece(0, 1, Bevel.LEFT, false);
			moveAPiece(0, 0, Bevel.BOTTOM, false);
			doubleRight = true;
		} else if (bevel.cubeLoc[0][5].face == bevel.cubeLoc[0][1].face) {
			moveAPiece(0, 0, Bevel.BOTTOM, false);
			right = true;
		}
		if (bevel.cubeLoc[1][4].face == bevel.cubeLoc[1][2].face) {
			moveAPiece(0, 3, Bevel.BOTTOM, false);
			moveAPiece(0, 0, Bevel.LEFT, false);
			doubleLeft = true;
		} else if (bevel.cubeLoc[0][7].face == bevel.cubeLoc[0][3].face) {
			moveAPiece(0, 0, Bevel.LEFT, false);
			left = true;
		}
		swap4(3, 0);
		if (doubleLeft) {
			moveAPiece(0, 0, Bevel.LEFT, false);
			moveAPiece(0, 3, Bevel.BOTTOM, false);
		} else if (left) {
			moveAPiece(0, 0, Bevel.LEFT, false);
		}
		if (doubleRight) {
			moveAPiece(0, 0, Bevel.BOTTOM, false);
			moveAPiece(0, 1, Bevel.LEFT, false);
		} else if (right) {
			moveAPiece(0, 0, Bevel.BOTTOM, false);
		}
	}

	void fixConnectedDouble() {
		boolean left = false, lowerLeft = false;
		boolean right = false, lowerRight = false;

		if (bevel.cubeLoc[0][6].face == bevel.cubeLoc[0][3].face) {
			moveAPiece(1, 3, Bevel.RIGHT, false);
			left = true;
		}
		if (left || bevel.cubeLoc[1][7].face == bevel.cubeLoc[1][2].face) {
			moveAPiece(2, 0, Bevel.BOTTOM, false);
			moveAPiece(3, 1, Bevel.LEFT, false);
			moveAPiece(4, 1, Bevel.LEFT, false);
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			lowerLeft = true;
		}
		if (bevel.cubeLoc[2][7].face == bevel.cubeLoc[2][2].face) {
			moveAPiece(3, 0, Bevel.BOTTOM, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(1, 2, Bevel.RIGHT, false);
			moveAPiece(1, 1, Bevel.TOP, false);
			right = true;
		}
		if (right || bevel.cubeLoc[3][4].face == bevel.cubeLoc[3][0].face) {
			moveAPiece(0, 1, Bevel.TOP, false);
			moveAPiece(3, 0, Bevel.BOTTOM, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(1, 2, Bevel.RIGHT, false);
			lowerRight = true;
		} else {
			moveAPiece(3, 0, Bevel.BOTTOM, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(1, 2, Bevel.RIGHT, false);
		}
		swap4(3, 1);
		if (lowerRight) {
			moveAPiece(1, 2, Bevel.RIGHT, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(3, 0, Bevel.BOTTOM, false);
			moveAPiece(0, 1, Bevel.TOP, false);
			if (right) {
				moveAPiece(1, 1, Bevel.TOP, false);
				moveAPiece(1, 2, Bevel.RIGHT, false);
				moveAPiece(5, 0, Bevel.LEFT, false);
				moveAPiece(3, 0, Bevel.BOTTOM, false);
			}
		} else {
			moveAPiece(1, 2, Bevel.RIGHT, false);
			moveAPiece(5, 0, Bevel.LEFT, false);
			moveAPiece(3, 0, Bevel.BOTTOM, false);
		}
		if (lowerLeft) {
			moveAPiece(5, 3, Bevel.BOTTOM, false);
			moveAPiece(4, 1, Bevel.LEFT, false);
			moveAPiece(3, 1, Bevel.LEFT, false);
			moveAPiece(2, 0, Bevel.BOTTOM, false);
			if (left) {
				moveAPiece(1, 3, Bevel.RIGHT, false);
			}
		}
	}

	/* This procedure coordinates the solution process. */
	void solvePieces() {
		if (!bevel.checkSolved()) {
			setSecondCorner();
			setThirdCorner();
			setFourthCorner();
			turnOver();
			setFifthCorner();
			setSixthCorner();
			setLastCorners();
			setFace0();
			setFaceTop();
			fixTriples();
			if (checkOddSplitDouble()) {
				fixOddSplitDouble();
			}
			if (checkEvenSplitDouble()) {
				boolean which = (generator.nextInt(2) != 0);
				moveAPiece(1, 0, ((which) ? Bevel.RIGHT : Bevel.LEFT), true);
				fixOddSplitDouble();
			}
			if (checkConnectedDouble()) {
				fixConnectedDouble();
			}
		}
		((BevelFrame) frame).callback(
			BevelInterface.ACTION_COMPUTED, bevel.dim);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solvePieces();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
